from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.boxes import (
    archive_boxes,
    bulk_get_boxes,
    create_box,
    get_box,
    list_box_contents,
    list_boxes,
    unarchive_boxes,
    update_box,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    Box,
    BoxContentsPaginatedList,
    BoxCreate,
    BoxesArchivalChange,
    BoxesArchive,
    BoxesArchiveReason,
    BoxesPaginatedList,
    BoxesUnarchive,
    BoxUpdate,
    ContainerWithCoordinates,
    ListBoxesSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class BoxService(BaseService):
    """
    Boxes.

    Boxes are a structured storage type, consisting of a grid of positions that can each hold one container.
    Unlike locations, there are a maximum number of containers that a box can hold (one per position).

    Boxes are all associated with schemas, which define the type of the box (e.g. "10x10 Cryo Box") along
    with the fields that are tracked and the dimensions of the box.

    Like all storage, every Box has a barcode that is unique across the registry.
    """

    @api_method
    def get_by_id(self, box_id: str) -> Box:
        """
        Get a box.

        See https://benchling.com/api/reference#/Boxes/getBox
        """
        response = get_box.sync_detailed(client=self.client, box_id=box_id)
        return model_from_detailed(response)

    @api_method
    def _boxes_page(
        self,
        *,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListBoxesSort] = None,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ancestor_storage_id: Optional[str] = None,
        storage_contents_id: Optional[str] = None,
        storage_contents_ids: Optional[List[str]] = None,
        empty_positions: Optional[int] = None,
        empty_positions_gte: Optional[int] = None,
        empty_positions_gt: Optional[int] = None,
        empty_positions_lte: Optional[int] = None,
        empty_positions_lt: Optional[int] = None,
        empty_containers: Optional[int] = None,
        empty_containers_gte: Optional[int] = None,
        empty_containers_gt: Optional[int] = None,
        empty_containers_lte: Optional[int] = None,
        empty_containers_lt: Optional[int] = None,
        ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
    ) -> Response[BoxesPaginatedList]:
        response = list_boxes.sync_detailed(
            client=self.client,
            sort=none_as_unset(sort),
            schema_id=none_as_unset(schema_id),
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            ancestor_storage_id=none_as_unset(ancestor_storage_id),
            storage_contents_id=none_as_unset(storage_contents_id),
            storage_contents_ids=none_as_unset(optional_array_query_param(storage_contents_ids)),
            empty_positions=none_as_unset(empty_positions),
            empty_positionsgte=none_as_unset(empty_positions_gte),
            empty_positionsgt=none_as_unset(empty_positions_gt),
            empty_positionslte=none_as_unset(empty_positions_lte),
            empty_positionslt=none_as_unset(empty_positions_lt),
            empty_containers=none_as_unset(empty_containers),
            empty_containersgte=none_as_unset(empty_containers_gte),
            empty_containersgt=none_as_unset(empty_containers_gt),
            empty_containerslte=none_as_unset(empty_containers_lte),
            empty_containerslt=none_as_unset(empty_containers_lt),
            ids=none_as_unset(optional_array_query_param(ids)),
            barcodes=none_as_unset(optional_array_query_param(barcodes)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            archive_reason=none_as_unset(archive_reason),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        sort: Optional[Union[str, ListBoxesSort]] = None,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ancestor_storage_id: Optional[str] = None,
        storage_contents_id: Optional[str] = None,
        storage_contents_ids: Optional[List[str]] = None,
        empty_positions: Optional[int] = None,
        empty_positions_gte: Optional[int] = None,
        empty_positions_gt: Optional[int] = None,
        empty_positions_lte: Optional[int] = None,
        empty_positions_lt: Optional[int] = None,
        empty_containers: Optional[int] = None,
        empty_containers_gte: Optional[int] = None,
        empty_containers_gt: Optional[int] = None,
        empty_containers_lte: Optional[int] = None,
        empty_containers_lt: Optional[int] = None,
        ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[Box]:
        """
        List boxes.

        See https://benchling.com/api/reference#/Boxes/listBoxes
        """
        check_for_csv_bug_fix("storage_contents_ids", storage_contents_ids)

        def api_call(next_token: NextToken) -> Response[BoxesPaginatedList]:
            return self._boxes_page(
                sort=_translate_to_string_enum(ListBoxesSort, sort),
                schema_id=schema_id,
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                ancestor_storage_id=ancestor_storage_id,
                storage_contents_id=storage_contents_id,
                storage_contents_ids=storage_contents_ids,
                empty_positions=empty_positions,
                empty_positions_gte=empty_positions_gte,
                empty_positions_gt=empty_positions_gt,
                empty_positions_lte=empty_positions_lte,
                empty_positions_lt=empty_positions_lt,
                empty_containers=empty_containers,
                empty_containers_gte=empty_containers_gte,
                empty_containers_gt=empty_containers_gt,
                empty_containers_lte=empty_containers_lte,
                empty_containers_lt=empty_containers_lt,
                ids=ids,
                barcodes=barcodes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                archive_reason=archive_reason,
                schema_fields=schema_fields,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: BoxesPaginatedList) -> Optional[List[Box]]:
            return body.boxes

        return PageIterator(api_call, results_extractor)

    @api_method
    def _box_contents_page(
        self,
        *,
        box_id: str,
        next_token: Optional[str] = None,
        page_size: Optional[int] = 50,
    ) -> Response[BoxContentsPaginatedList]:
        response = list_box_contents.sync_detailed(
            client=self.client,
            box_id=box_id,
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_box_contents(
        self,
        *,
        box_id: str,
        page_size: Optional[int] = 50,
    ) -> PageIterator[ContainerWithCoordinates]:
        """
        List a box's contents.

        See https://benchling.com/api/reference#/Boxes/listBoxContents
        """

        def api_call(next_token: NextToken) -> Response[BoxContentsPaginatedList]:
            return self._box_contents_page(
                box_id=box_id,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: BoxContentsPaginatedList) -> Optional[List[ContainerWithCoordinates]]:
            return body.containers

        return PageIterator(api_call, results_extractor)

    @api_method
    def bulk_get(
        self, *, box_ids: Optional[Iterable[str]] = None, barcodes: Optional[Iterable[str]] = None
    ) -> Optional[List[Box]]:
        """
        Bulk get boxes.

        See https://benchling.com/api/reference#/Boxes/bulkGetBoxes
        """
        box_id_string = optional_array_query_param(box_ids)
        barcode_string = optional_array_query_param(barcodes)
        response = bulk_get_boxes.sync_detailed(
            client=self.client, box_ids=none_as_unset(box_id_string), barcodes=none_as_unset(barcode_string)
        )
        boxes_list = model_from_detailed(response)
        return boxes_list.boxes

    @api_method
    def create(self, box: BoxCreate) -> Box:
        """
        Create a box.

        See https://benchling.com/api/reference#/Boxes/createBox
        """
        response = create_box.sync_detailed(client=self.client, json_body=box)
        return model_from_detailed(response)

    @api_method
    def update(self, box_id: str, box: BoxUpdate) -> Box:
        """
        Update a box.

        See https://benchling.com/api/reference#/Boxes/updateBox
        """
        response = update_box.sync_detailed(client=self.client, box_id=box_id, json_body=box)
        return model_from_detailed(response)

    @api_method
    def archive(
        self, box_ids: Iterable[str], reason: BoxesArchiveReason, should_remove_barcodes: bool
    ) -> BoxesArchivalChange:
        """
        Archive boxes and any containers of the boxes.

        See https://benchling.com/api/reference#/Boxes/archiveBoxes
        """
        archive_request = BoxesArchive(
            box_ids=list(box_ids), reason=reason, should_remove_barcodes=should_remove_barcodes
        )
        response = archive_boxes.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, box_ids: Iterable[str]) -> BoxesArchivalChange:
        """
        Unarchive boxes and the containers that were archived along with them.

        See https://benchling.com/api/reference#/Boxes/unarchiveBoxes
        """
        unarchive_request = BoxesUnarchive(box_ids=list(box_ids))
        response = unarchive_boxes.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)



import base64
from datetime import datetime, timedelta, timezone
from typing import cast, List

from benchling_api_client.v2.benchling_client import BenchlingApiClient
import httpx
from jwcrypto import jwk
from typing_extensions import Protocol

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.logging_helpers import log_stability_warning, StabilityLevel
from benchling_sdk.helpers.package_helpers import _required_packages_context, ExtrasPackage
from benchling_sdk.services.v2.stable.api_service import build_json_response

log_stability_warning(StabilityLevel.ALPHA)


class WebhookVerificationError(Exception):
    """
    Webhook Verification Error.

    Indicates a webhook from Benchling could not be verified.

    Some reasons this could happen include:
    - An app developer misconfiguration (e.g., wrong app)
    - A webhook was received too late
    - The webhook did not originate from Benchling (possible attack vector)
    """

    pass


class GetJwksFunction(Protocol):
    """Function for custom resolution of JWKs for an app."""

    def __call__(self, app_id: str) -> jwk.JWKSet:
        """Retrieve JWKs for an app."""
        pass


class JwkUrlProvider(Protocol):
    """
    Function for custom URL for resolving JWKs by app.

    Should generally never be used except in the event of specialized testing with internal Benchling.
    """

    def __call__(self, app_id: str) -> str:
        """Return a fully qualified URL for retrieving JWKs from an app."""
        pass


def _default_httpx_webhook_client() -> httpx.Client:
    """
    Create a default httpx client for webhook verification.

    Since webhook verification does not require authentication and typically takes place before
    Benchling initialization, we don't have access to the internal httpx in Benchling.
    """
    transport = httpx.HTTPTransport(retries=3)
    # noinspection PyProtectedMember
    user_agent_header = BenchlingApiClient._get_user_agent("Benchling SDK", "benchling-sdk")
    return httpx.Client(transport=transport, headers={"User-Agent": user_agent_header})


def _production_webhook_jwks(app_id) -> str:
    """Return the URL for retrieving production JWKs for an app."""
    return f"https://benchling.com/apps/jwks/{app_id}"


def jwks_by_app(
    app_id: str,
    httpx_client: httpx.Client = None,
    jwk_url_provider: JwkUrlProvider = None,
) -> jwk.JWKSet:
    """
    Get JWKs by App.

    Retrieves a set of JWKs assigned to an app used to verify webhooks.

    JWKs generally should not be resolved on their own. We recommend using webhook verification
    functions such as verify().

    This method is provided for specialized cases such as customizing the httpx client.
    """
    if httpx_client is None:
        httpx_client = _default_httpx_webhook_client()
    if jwk_url_provider is None:
        jwk_url_provider = _production_webhook_jwks
    jwk_url = jwk_url_provider(app_id)
    httpx_response = httpx_client.get(jwk_url)
    response = build_json_response(httpx_response)
    raise_for_status(response)
    return jwk.JWKSet.from_json(response.content)


def verify(app_id: str, data: str, headers: dict, jwk_function: GetJwksFunction = None) -> None:
    """
    Verify a webhook.

    Verifies that a webhook was a valid webhook from Benchling.
    Raises WebhookVerificationError if the webhook could not be verified.
    Resolves JWKs from Benchling with default settings. Pass jwk_function for customization.
    """
    _verify_headers_present(headers)
    _verify_timestamp(headers["webhook-timestamp"])
    to_verify = f'{headers["webhook-id"]}.{headers["webhook-timestamp"]}.{data}'
    signatures = headers["webhook-signature"].split(" ")
    der_signatures = _der_signatures_from_versioned_signatures(signatures)
    base64_der_signatures = [base64.b64decode(der_signature) for der_signature in der_signatures]
    if jwk_function is None:

        def _get_default_jwks(_app_id):
            return jwks_by_app(_app_id)

        jwk_function = _get_default_jwks
    jwks = jwk_function(app_id)
    if not _has_valid_signature(to_verify, jwks, base64_der_signatures):
        raise WebhookVerificationError("No matching signature found")


@_required_packages_context(ExtrasPackage.CRYPTOGRAPHY)
def _has_valid_signature(to_verify: str, jwks: jwk.JWKSet, encoded_signatures: List[bytes]) -> bool:
    from cryptography.exceptions import InvalidSignature
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.asymmetric import ec
    from cryptography.hazmat.primitives.serialization import load_pem_public_key

    for jk in jwks:
        pubkey = cast(ec.EllipticCurvePublicKey, load_pem_public_key(jk.export_to_pem()))
        for der_signature in encoded_signatures:
            try:
                pubkey.verify(der_signature, bytes(to_verify, "utf-8"), ec.ECDSA(hashes.SHA256()))
            except InvalidSignature:
                continue
            return True
    return False


def _der_signatures_from_versioned_signatures(versioned_signatures: str) -> List[str]:
    """
    Parse and return a list of ders signatures from a single string.

    Signature format is f"v{version_number}der,{signature}"
    """
    der_signatures = []
    for versioned_sig in versioned_signatures:
        _version, sig = versioned_sig.split(",")
        if "der" in _version:
            der_signatures.append(sig)
    if len(der_signatures) == 0:
        raise WebhookVerificationError("Expected to find a der encoded signature")
    return der_signatures


def _verify_headers_present(headers: dict) -> None:
    """
    Verify Headers Present.

    Check that webhook headers contain all the expected keys. Raises WebhookVerificationError if
    headers are missing, which is more friendly than KeyError.

    Does not validate the contents of headers.
    """
    if "webhook-id" not in headers:
        raise WebhookVerificationError("Missing webhook-id header")
    if "webhook-timestamp" not in headers:
        raise WebhookVerificationError("Missing webhook-timestamp header")
    if "webhook-signature" not in headers:
        raise WebhookVerificationError("Missing webhook-signature header")


def _verify_timestamp(timestamp_header: str) -> None:
    """
    Verify Timestamp.

    Checks that a timestamp in the webhook header is recent enough to not suspect a replay attack.
    """
    webhook_tolerance = timedelta(minutes=5)
    now = datetime.now(tz=timezone.utc)
    try:
        timestamp = datetime.fromtimestamp(float(timestamp_header), tz=timezone.utc)
    except Exception:
        raise WebhookVerificationError("Invalid Signature Headers")
    if timestamp < (now - webhook_tolerance):
        raise WebhookVerificationError("Message timestamp too old")
    if timestamp > (now + webhook_tolerance):
        raise WebhookVerificationError("Message timestamp too new")



from typing import Iterable, Optional, Type

from benchling_api_client.v2.types import Response

from benchling_sdk.errors import ExtendedBenchlingErrorBase, raise_for_status


def model_from_detailed(
    response: Response, error_types: Optional[Iterable[Type[ExtendedBenchlingErrorBase]]] = None
):
    """
    Deserialize a response into a model.

    May optionally take error_types which can produce an error_matcher() for more specific error cases.
    """
    matchers = [e.error_matcher() for e in error_types] if error_types else None
    raise_for_status(response, matchers)
    return response.parsed



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import cast, Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union

from benchling_api_client.v2.extensions import UnknownType
from ordered_set import OrderedSet
from typing_extensions import Protocol

from benchling_sdk.apps.config.decryption_provider import BaseDecryptionProvider
from benchling_sdk.apps.config.scalars import (
    DEFAULT_SCALAR_DEFINITIONS,
    scalar_definition_from_field_type,
    ScalarConfigItemType,
    ScalarDefinition,
    ScalarType,
)
from benchling_sdk.benchling import Benchling
from benchling_sdk.models import (
    AppConfigItem,
    ArrayElementAppConfigItem,
    BooleanAppConfigItem,
    DateAppConfigItem,
    DatetimeAppConfigItem,
    EntitySchemaAppConfigItem,
    Field,
    FieldAppConfigItem,
    FloatAppConfigItem,
    GenericApiIdentifiedAppConfigItem,
    InaccessibleResource,
    IntegerAppConfigItem,
    JsonAppConfigItem,
    LinkedAppConfigResourceSummary,
    ListAppConfigurationItemsSort,
    SecureTextAppConfigItem,
    TextAppConfigItem,
)


class MissingDependencyError(Exception):
    """
    Missing dependency error.

    Indicates a dependency was missing from app config.
    For instance, no dependency with that name was in the list.
    """

    pass


class UnsupportedDependencyError(Exception):
    """
    Unsupported dependency error.

    The manifest and configuration specified a dependency which the SDK is incapable of handling yet.
    """

    pass


class MissingScalarDefinitionError(Exception):
    """
    Missing scalar definition error.

    The manifest and configuration specified a scalar type which the SDK does not know how to translate
    to Python values yet.
    """

    pass


class InaccessibleAppConfigResourceError(Exception):
    """
    Inaccessible app config resource error.

    A resource was linked in app config, but the permissions do not allow access to it. Most likely happens
    when an app lacks the necessary permissions.
    """

    pass


ConfigItemPath = Tuple[str, ...]

# Everything from AppConfigItem except UnknownType
ConfigurationReference = Union[
    ArrayElementAppConfigItem,
    DateAppConfigItem,
    DatetimeAppConfigItem,
    JsonAppConfigItem,
    EntitySchemaAppConfigItem,
    FieldAppConfigItem,
    BooleanAppConfigItem,
    IntegerAppConfigItem,
    FloatAppConfigItem,
    GenericApiIdentifiedAppConfigItem,
    SecureTextAppConfigItem,
    TextAppConfigItem,
]

ConfigWithLinkedResource = Union[
    EntitySchemaAppConfigItem,
    FieldAppConfigItem,
    GenericApiIdentifiedAppConfigItem,
]

ScalarConfigReference = Union[
    BooleanAppConfigItem,
    DateAppConfigItem,
    DatetimeAppConfigItem,
    FloatAppConfigItem,
    IntegerAppConfigItem,
    JsonAppConfigItem,
    SecureTextAppConfigItem,
    TextAppConfigItem,
]

D = TypeVar("D", bound="BaseDependencies")
F = TypeVar("F", bound="BaseField")


class ConfigProvider(Protocol):
    """
    Config provider.

    Provides a BenchlingAppConfiguration.
    """

    def config(self) -> List[ConfigurationReference]:
        """Implement to provide a Benchling app configuration."""
        pass


class BenchlingConfigProvider(ConfigProvider):
    """
    Benchling Config provider.

    Provides a BenchlingAppConfiguration retrieved from Benchling's API.
    """

    _client: Benchling
    _app_id: str

    def __init__(self, client: Benchling, app_id: str):
        """
        Initialize Benchling Config Provider.

        :param client: A configured Benchling instance for making API calls.
        :param app_id: The app_id from which to retrieve configuration.
        """
        self._client = client
        self._app_id = app_id

    def config(self) -> List[ConfigurationReference]:
        """Provide a Benchling app configuration from Benchling's APIs."""
        app_pages = self._client.apps.list_app_configuration_items(
            app_id=self._app_id,
            page_size=100,
            sort=ListAppConfigurationItemsSort.CREATEDATASC,
        )

        # Eager load all config items for now since we don't yet have a way of lazily querying by path
        all_config_pages = list(app_pages)
        # Punt on UnknownType for now as apps using manifests with new types could lead to unpredictable results
        all_config_items = [
            _supported_config_item(config_item) for page in all_config_pages for config_item in page
        ]

        return all_config_items


class StaticConfigProvider(ConfigProvider):
    """
    Static Config provider.

    Provides a BenchlingAppConfiguration from a static declaration. Useful for mocking or testing.
    """

    _configuration_items: List[ConfigurationReference]

    def __init__(self, configuration_items: List[ConfigurationReference]):
        """
        Initialize Static Config Provider.

        :param configuration_items: The configuration items to return.
        """
        self._configuration_items = configuration_items

    def config(self) -> List[ConfigurationReference]:
        """Provide Benchling app configuration items from a static reference."""
        return self._configuration_items


class DependencyLinkStore(object):
    """
    Dependency Link Store.

    Marshalls an app configuration from the configuration provider into an indexable structure.
    Only retrieves app configuration once unless its cache is invalidated.
    """

    _configuration_provider: ConfigProvider
    _configuration: Optional[List[ConfigurationReference]] = None
    _configuration_map: Optional[Dict[ConfigItemPath, ConfigurationReference]] = None
    _array_path_row_names: Dict[Tuple[str, ...], OrderedSet[str]] = dict()

    def __init__(self, configuration_provider: ConfigProvider):
        """
        Initialize Dependency Link Store.

        :param configuration_provider: A ConfigProvider that will be invoked to provide the
        underlying config from which to organize dependency links.
        """
        self._configuration_provider = configuration_provider
        self._array_path_row_names = dict()

    @classmethod
    def from_app(cls, client: Benchling, app_id: str) -> DependencyLinkStore:
        """
        From App.

        Instantiate a DependencyLinkStore from an app_id and a configured Benchling instance. Preferred to
        using the class's constructor.
        """
        config_provider = BenchlingConfigProvider(client, app_id)
        return cls(config_provider)

    @property
    def configuration(self) -> List[ConfigurationReference]:
        """
        Get the underlying configuration.

        Return the raw, stored configuration. Can be used if the provided accessors are inadequate
        to find particular configuration items.
        """
        if not self._configuration:
            self._configuration = self._configuration_provider.config()
        return self._configuration

    @property
    def configuration_path_map(self) -> Dict[ConfigItemPath, ConfigurationReference]:
        """
        Config links.

        Return a map of configuration item paths to their corresponding configuration items.
        """
        if not self._configuration_map:
            self._configuration_map = {tuple(item.path): item for item in self.configuration}
        return self._configuration_map

    def config_by_path(self, path: List[str]) -> Optional[ConfigurationReference]:
        """
        Config by path.

        Find an app config item by its exact path match, if it exists. Does not search partial paths.
        """
        # Since we eager load all config now, we know that missing path means it's not configured in Benchling
        # Later if we support lazy loading, we'll need to differentiate what's in our cache versus missing
        return self.configuration_path_map.get(tuple(path))

    def config_keys_by_path(self, path: List[str]) -> OrderedSet[str]:
        """
        Config keys by path.

        Find a set of app config keys at the specified path, if any. Does not return keys that are nested
        beyond the current level.

        For instance, given paths:
        ["One", "Two"]
        ["One", "Two", "Three"]
        ["One", "Two", "Four"]
        ["One", "Two", "Three", "Five"]
        ["Zero", "One", "Two", "Three"]

        The expected return from this method when path=["One", "Two"] is a set {"Three", "Four"}.
        """
        # Convert path to tuple, as list is not hashable for dict keys
        path_tuple = tuple(path)
        if path_tuple not in self._array_path_row_names:
            self._array_path_row_names[path_tuple] = OrderedSet(
                [
                    config_item.path[len(path)]
                    # Use the list instead of configuration_map to preserve order
                    for config_item in self.configuration
                    # The +1 is the name of the array row
                    if len(config_item.path) >= len(path) + 1
                    # Ignoring flake8 error E203 because black keeps putting in whitespace padding :
                    and config_item.path[0 : len(path_tuple)] == path  # noqa: E203
                    and config_item.value is not None
                ]
            )
        return self._array_path_row_names[path_tuple]

    def invalidate_cache(self) -> None:
        """
        Invalidate Cache.

        Will force retrieval of configuration from the ConfigProvider the next time the link store is accessed.
        """
        self._configuration = None
        self._configuration_map = None
        self._array_path_row_names = dict()


class HasAppConfigItem(Protocol):
    """
    Has App Config Item.

    A mixin for typing to assert that a class has an optional app config item attribute.
    """

    @property
    def path(self) -> List[str]:
        """Return the path requested by the manifest."""
        pass

    @property
    def config_item(self) -> Optional[ConfigurationReference]:
        """Return the underlying app config item, if present."""
        pass


class HasApiIdentifiedAppConfigItem(Protocol):
    """
    Has Api Identified App Config Item.

    A mixin for typing to assert that a class has an optional app config item attribute.
    That app config item must have a linked_resource property.
    """

    @property
    def path(self) -> List[str]:
        """Return the path requested by the manifest."""
        pass

    @property
    def config_item(self) -> Optional[ConfigWithLinkedResource]:
        """Return the underlying app config item, if present. App config item must have linked_resource."""
        pass


class HasScalarDefinition(Protocol):
    """
    Has Scalar Definition.

    A mixin for typing to assert that a particular class has scalar attributes.
    """

    @property
    def path(self) -> List[str]:
        """Return the path requested by the manifest."""
        pass

    @property
    def config_item(self) -> Optional[ConfigurationReference]:
        """Return the underlying app config item, if present."""
        pass

    @property
    def definition(self) -> Optional[ScalarDefinition]:
        """Return the scalar definition, allowing for conversion to Python types."""
        pass


class HasConfigWithDecryptionProvider(Protocol):
    """
    Has Config With Decryption Provider.

    A mixin for typing to assert that a particular class has a decryption provider and config.
    """

    @property
    def path(self) -> List[str]:
        """Return the path requested by the manifest."""
        pass

    @property
    def config_item(self) -> Optional[ConfigurationReference]:
        """Return the underlying app config item, if present."""
        pass

    @property
    def decryption_provider(self) -> Optional[BaseDecryptionProvider]:
        """Return the decryption provider."""
        pass


class RequiredLinkedResourceDependencyMixin:
    """
    Required Linked Resource Dependency Mixin.

    A mixin for easily accessing attributes from linked_resource for an app config item which is required and
    should always be present. Should only be mixed in with HasApiIdentifiedAppConfigItem
    """

    @property
    def id(self: HasApiIdentifiedAppConfigItem) -> str:
        """Return the API ID of the linked configuration."""
        assert (
            self.config_item is not None and self.config_item.value is not None
        ), f"The app config item {self.path} is not set in Benchling"
        # config_item.value and linked_resource.id are the same for now,
        # so we can eschew inaccessible resource checking
        return self.config_item.value

    @property
    def name(self: HasApiIdentifiedAppConfigItem) -> str:
        """Return the name of the linked configuration.

        Raises InaccessibleAppConfigResourceError if the app does not have permission to the linked resource.
        """
        assert (
            self.config_item is not None and self.config_item.value is not None
        ), f"The app config item {self.path} is not set in Benchling"
        if isinstance(self.config_item.linked_resource, InaccessibleResource):
            raise InaccessibleAppConfigResourceError(
                f'No permissions to the linked resource "{self.config_item.value}" referenced by {self.path}'
            )
        # Required for type checking
        assert isinstance(
            self.config_item.linked_resource, LinkedAppConfigResourceSummary
        ), f"Expected linked resource from app config item but got {type(self.config_item.linked_resource)}"
        return self.config_item.linked_resource.name


class OptionalLinkedResourceDependencyMixin:
    """
    Optional Linked Resource Dependency Mixin.

    A mixin for easily accessing attributes from linked_resource for an app config item which is optional and
    may not be present. Should only be mixed in with HasApiIdentifiedAppConfigItem
    """

    @property
    def id(self: HasApiIdentifiedAppConfigItem) -> Optional[str]:
        """Return the API ID of the linked configuration, if present."""
        # config_item.value and linked_resource.id are the same for now,
        # so we can eschew inaccessible resource checking
        if self.config_item is not None and self.config_item.value is not None:
            return self.config_item.value
        return None

    @property
    def name(self: HasApiIdentifiedAppConfigItem) -> Optional[str]:
        """Return the name of the linked configuration, if present.

        Raises InaccessibleAppConfigResourceError if the app does not have permission to the linked resource.
        """
        if self.config_item is not None and self.config_item.value is not None:
            if isinstance(self.config_item.linked_resource, InaccessibleResource):
                raise InaccessibleAppConfigResourceError(
                    f'No permissions to the linked resource "{self.config_item.value}" referenced by {self.path}'
                )
            # Required for type checking
            assert isinstance(
                self.config_item.linked_resource, LinkedAppConfigResourceSummary
            ), f"Expected linked resource from app config item but got {type(self.config_item.linked_resource)}"
            return self.config_item.linked_resource.name
        return None


class OptionalValueMixin:
    """
    Optional Value Mixin.

    A mixin for accessing a value which is optional and may not be present. Should
    only be mixed in with HasAppConfigItem or another class that provides the `self.config_item` attribute.
    """

    @property
    def value(self: HasAppConfigItem) -> Optional[str]:
        """Return the value of the app config item, if present."""
        if self.config_item and self.config_item.value:
            return str(self.config_item.value)
        return None


class RequiredValueMixin:
    """
    Required Value Mixin.

    A mixin for accessing a value which is required and should always be present. Should
    only be mixed in with HasAppConfigItem or another class that provides the `self.config_item` attribute.
    """

    @property
    def value(self: HasAppConfigItem) -> str:
        """Return the value of the app config item."""
        assert (
            self.config_item is not None and self.config_item.value is not None
        ), f"The app config item {self.path} is not set in Benchling"
        return str(self.config_item.value)


class RequiredScalarDependencyMixin(Generic[ScalarType]):
    """
    Require Scalar Config.

    A mixin for accessing a scalar config which is required and should always be present.
    Should only be mixed in with HasScalarDefinition.
    """

    @property
    def value(self: HasScalarDefinition) -> ScalarType:
        """Return the value of the scalar."""
        if self.definition:
            assert (
                self.config_item is not None and self.config_item.value is not None
            ), f"The app config item {self.path} is not set in Benchling"
            optional_typed_value = self.definition.from_str(value=str(self.config_item.value))
            assert optional_typed_value is not None
            return optional_typed_value
        raise MissingScalarDefinitionError(f"No definition registered for scalar config {self.path}")

    @property
    def value_str(self: HasScalarDefinition) -> str:
        """Return the value of the scalar as a string."""
        assert (
            self.config_item is not None and self.config_item.value is not None
        ), f"The app config item {self.path} is not set in Benchling"
        # Booleans are currently specified as str in the spec but are bool at runtime in JSON
        return str(self.config_item.value)


class OptionalScalarDependencyMixin(Generic[ScalarType]):
    """
    Optional Scalar Config.

    A mixin for accessing a scalar config which is optional and may not be present.
    Should only be mixed in with HasScalarDefinition.
    """

    @property
    def value(self: HasScalarDefinition) -> Optional[ScalarType]:
        """Return the value of the scalar, if present."""
        if self.config_item and self.config_item.value:
            if self.definition:
                return self.definition.from_str(value=str(self.config_item.value))
            raise MissingScalarDefinitionError(f"No definition registered for scalar config {self.path}")
        return None

    @property
    def value_str(self: HasScalarDefinition) -> Optional[str]:
        """Return the value of the scalar as a string, if present."""
        if self.config_item and self.config_item.value:
            return str(self.config_item.value)
        return None


class RequiredSecureTextDependencyMixin(RequiredScalarDependencyMixin[str]):
    """
    Require Secure Text.

    A mixin for accessing a secure text config which is required and should always be present.
    Should only be mixed in with SecureTextConfig.
    """

    def decrypted_value(self: HasConfigWithDecryptionProvider) -> str:
        """
        Decrypted value.

        Decrypts a secure_text dependency's encrypted value into plain text.
        """
        assert (
            self.config_item is not None and self.config_item.value is not None
        ), f"The app config item {self.path} is not set in Benchling"
        assert (
            self.decryption_provider is not None
        ), f"The app config item {self.config_item} cannot be decrypted because no DecryptionProvider was set"
        return self.decryption_provider.decrypt(str(self.config_item.value))


class OptionalSecureTextDependencyMixin(OptionalScalarDependencyMixin[str]):
    """
    Optional Secure Text.

    A mixin for accessing a secure text config which is optional and may not be present.
    Should only be mixed in with SecureTextConfig.
    """

    def decrypted_value(self: HasConfigWithDecryptionProvider) -> Optional[str]:
        """
        Decrypted value.

        Decrypts a secure_text dependency's encrypted value into plain text, if present.
        """
        if self.config_item and self.config_item.value:
            assert (
                self.decryption_provider is not None
            ), f"The app config item {self.config_item} cannot be decrypted because no DecryptionProvider was set"
            return self.decryption_provider.decrypt(str(self.config_item.value))
        return None


@dataclass
class BaseConfigNode:
    """
    Base Config Node.

    A node in a graph of related config items, referencing its parent.

    All nodes should have a parent, which may be BaseDependencies, but not all nodes represent an AppConfigItem
    in Benchling.
    """

    parent: Union[BaseDependencies, BaseConfigNode]

    def context(self) -> BaseDependencies:
        """Return the dependency class at the root of the dependency graph."""
        if isinstance(self.parent, BaseDependencies):
            return self.parent
        return self.parent.context()

    def full_path(self) -> List[str]:
        """Return the full path of the current node, inheriting from all parents."""
        parent_path = self.parent.full_path() if isinstance(self.parent, BaseConfigNode) else []
        # Fields and options classes typically don't define paths
        last_path = getattr(self, "path", [])
        return parent_path + last_path


@dataclass
class BaseConfigItem(BaseConfigNode):
    """
    Base Config Item.

    A reference to any config item.
    """

    config_item: Optional[ConfigurationReference]


@dataclass
class ApiConfigItem(BaseConfigItem):
    """
    API Config Item.

    A reference to a config item for a Benchling object referencable by the API.
    """

    config_item: Optional[ConfigWithLinkedResource]


class BaseField(ABC, Field, Generic[ScalarType]):
    """
    Base Field.

    Provides additional accessors on top of the OpenAPI Field model.
    """

    @classmethod
    def from_field(cls: Type[F], field: Optional[Field]) -> F:
        """
        From Field.

        Create a new instance from an existing field.
        """
        if field:
            return cls(
                value=field.value,
                display_value=field.display_value,
                is_multi=field.is_multi,
                text_value=field.text_value,
                type=field.type,
            )
        return cls(value=None)

    @property
    def scalar_definition(self) -> ScalarDefinition[ScalarType]:
        """
        Scalar Definition.

        Returns a scalar definition for parsing a concrete type from a field.
        Override to implement custom deserialization.
        """
        return scalar_definition_from_field_type(self.type)


class RequiredField(BaseField, Generic[ScalarType]):
    """
    Required Field.

    A decorator class providing typed accessors for an underlying Field.
    Use with required, single valued fields.
    """

    @dataclass
    class _RequiredFieldTyped:
        base_field: RequiredField

        @property
        def value(self) -> ScalarType:
            """
            Typed Value.

            Returns the value of the field typed as it's specified in an app manifest.
            """
            # Can be None in the case of someone changing config
            typed_value = self.base_field.scalar_definition.from_str(value=str(self.base_field.value))
            assert typed_value is not None
            return typed_value

        @property
        def display_value(self) -> str:
            """
            Display Value.

            Return the field's display value as a string.
            """
            assert self.base_field.display_value is not None
            return self.base_field.display_value

    @property
    def typed(self) -> _RequiredFieldTyped:
        """
        Typed.

        Return a reference to a typed field with typesafe accessors.
        """
        return self._RequiredFieldTyped(self)


class OptionalField(BaseField, Generic[ScalarType]):
    """
    Optional Field.

    A decorator class providing typed accessors for an underlying Field.
    Use with optional, single valued fields.
    """

    @dataclass
    class _OptionalFieldTyped:
        base_field: OptionalField

        @property
        def value(self) -> Optional[ScalarType]:
            """
            Typed Value.

            Returns the value of the field typed as it's specified in an app manifest, if the field is present.
            """
            if self.base_field.value:
                field_value = str(self.base_field.value)
                return self.base_field.scalar_definition.from_str(value=field_value)
            return None

        @property
        def display_value(self) -> Optional[str]:
            """
            Display Value.

            Return the field's display value as a string, if present.
            """
            # Check to ensure linked before display_value, which will raise NotPresentError on unlinked
            if self.base_field.value is None:
                return None
            return None if self.base_field.display_value == "" else self.base_field.display_value

    @property
    def typed(self) -> _OptionalFieldTyped:
        """
        Typed.

        Return a reference to a typed field with typesafe accessors.
        """
        return self._OptionalFieldTyped(self)


class RequiredMultiValueField(BaseField, Generic[ScalarType]):
    """
    Required Multi Value Field.

    A decorator class providing typed accessors for an underlying Field.
    Use with required, multi-valued fields.
    """

    @dataclass
    class _RequiredMultiValueFieldType:
        base_field: RequiredMultiValueField

        @property
        def value(self) -> List[ScalarType]:
            """
            Typed Value.

            Returns the list of values in the field typed as it's specified in an app manifest.
            """
            typed_values: List[ScalarType] = [
                cast(ScalarType, self.base_field.scalar_definition.from_str(value=str(field_value)))
                for field_value in cast(List[str], self.base_field.value)
                if field_value is not None
            ]
            return typed_values

        @property
        def display_value(self) -> str:
            """
            Display Value.

            Return the field's display value as a string.
            """
            # We could try to return display value as List[str] for multi-valued fields, except comma is
            # an unreliable delimiter since the names within each value can contain it
            assert self.base_field.display_value is not None
            return self.base_field.display_value

    @property
    def typed(self) -> _RequiredMultiValueFieldType:
        """
        Typed.

        Return a reference to a typed field with typesafe accessors.
        """
        return self._RequiredMultiValueFieldType(self)


class OptionalMultiValueField(BaseField, Generic[ScalarType]):
    """
    Optional Multi Value Field.

    A decorator class providing typed accessors for an underlying Field.
    Use with optional, multi-valued fields.
    """

    @dataclass
    class _OptionalMultiValueFieldType:
        base_field: OptionalMultiValueField

        @property
        def value(self) -> Optional[List[ScalarType]]:
            """
            Typed Value.

            Returns the list of values in the field typed as it's specified in an app manifest, if present.
            """
            if self.base_field.value:
                typed_values: List[ScalarType] = [
                    cast(ScalarType, self.base_field.scalar_definition.from_str(value=str(field_value)))
                    for field_value in cast(List[str], self.base_field.value)
                    if field_value is not None
                ]
                return typed_values
            return None

        @property
        def display_value(self) -> Optional[str]:
            """
            Display Value.

            Return the field's display value as a string, if present.
            """
            # Check to ensure linked before display_value, which will raise NotPresentError on unlinked
            if self.base_field.value is None:
                return None
            return None if self.base_field.display_value == "" else self.base_field.display_value

    @property
    def typed(self) -> _OptionalMultiValueFieldType:
        """
        Typed.

        Return a reference to a typed field with typesafe accessors.
        """
        return self._OptionalMultiValueFieldType(self)


class RequiredSingleOrMultiValueField(BaseField, Generic[ScalarType]):
    """
    Required Single Or Multi Value Field.

    A decorator class providing typed accessors for an underlying Field.
    Use with required fields where isMulti is unset.
    """

    @dataclass
    class _RequiredSingleOrMultiValueFieldTyped:
        base_field: RequiredSingleOrMultiValueField

        @property
        def value(self) -> Union[ScalarType, List[ScalarType]]:
            """
            Typed Value.

            Returns the value in the field typed as it's specified in an app manifest.
            """
            if isinstance(self.base_field.value, list):
                typed_values: List[ScalarType] = [
                    cast(ScalarType, self.base_field.scalar_definition.from_str(value=str(field_value)))
                    for field_value in cast(List[str], self.base_field.value)
                    if field_value is not None
                ]
                return typed_values
            field_value = str(self.base_field.value)
            return cast(ScalarType, self.base_field.scalar_definition.from_str(value=field_value))

        @property
        def display_value(self) -> str:
            """
            Display Value.

            Return the field's display value as a string.
            """
            assert self.base_field.display_value is not None
            return self.base_field.display_value

    @property
    def typed(self) -> _RequiredSingleOrMultiValueFieldTyped:
        """
        Typed.

        Return a reference to a typed field with typesafe accessors.
        """
        return self._RequiredSingleOrMultiValueFieldTyped(self)


class OptionalSingleOrMultiValueField(BaseField, Generic[ScalarType]):
    """
    Optional Single Or Multi Value Field.

    A decorator class providing typed accessors for an underlying Field.
    Use with optional fields where isMulti is unset.
    """

    @dataclass
    class _OptionalSingleOrMultiValueFieldTyped:
        base_field: OptionalSingleOrMultiValueField

        @property
        def value(self) -> Optional[Union[ScalarType, List[ScalarType]]]:
            """
            Typed Value.

            Returns the value in the field typed as it's specified in an app manifest, if present.
            """
            if self.base_field.value is not None:
                if isinstance(self.base_field.value, list):
                    typed_values: List[ScalarType] = [
                        cast(ScalarType, self.base_field.scalar_definition.from_str(value=str(field_value)))
                        for field_value in cast(List[str], self.base_field.value)
                        if field_value is not None
                    ]
                    return typed_values
                else:
                    field_value = str(self.base_field.value)
                    return self.base_field.scalar_definition.from_str(value=field_value)
            return None

        @property
        def display_value(self) -> Optional[str]:
            """
            Display Value.

            Return the field's display value as a string, if present.
            """
            # Check to ensure linked before display_value, which will raise NotPresentError on unlinked
            if self.base_field.value is None:
                return None
            return None if self.base_field.display_value == "" else self.base_field.display_value

    @property
    def typed(self) -> _OptionalSingleOrMultiValueFieldTyped:
        """
        Typed.

        Return a reference to a typed field with typesafe accessors.
        """
        return self._OptionalSingleOrMultiValueFieldTyped(self)


@dataclass
class ArrayConfigItem(BaseConfigItem):
    """
    Array Config Item.

    An array config item representing a row.
    """

    @property
    def name(self) -> str:
        """Return the user defined name of the array row."""
        # Config item is not optional for arrays
        assert self.config_item is not None
        assert self.config_item.value is not None
        return cast(str, self.config_item.value)

    def full_path(self) -> List[str]:
        """Return the full path of the current array row, inheriting from all parents."""
        path = super().full_path()
        return path + [self.name]


@dataclass
class ScalarConfigItem(BaseConfigItem):
    """
    Scalar Config Item.

    Scalars are values that can be represented outside the Benchling domain.
    """

    config_item: Optional[ScalarConfigReference]
    definition: Optional[ScalarDefinition]


@dataclass
class SecureTextDependency(ScalarConfigItem):
    """
    SecureText Config.

    A dependency for accessing a secure_text config.
    """

    # This is declared Optional because a decryption provider is not required until attempting
    # to decrypt a value.
    decryption_provider: Optional[BaseDecryptionProvider]


class BaseDependencies:
    """
    A base class for implementing dependencies.

    Used as a facade for the underlying link store, which holds dependency links configured in Benchling.
    """

    _store: DependencyLinkStore
    _scalar_definitions: Dict[ScalarConfigItemType, ScalarDefinition]
    _unknown_scalar_definition: Optional[ScalarDefinition]
    # Will be required at runtime if an app attempts to decrypt a secure_text config
    _decryption_provider: Optional[BaseDecryptionProvider]

    def __init__(
        self,
        store: DependencyLinkStore,
        scalar_definitions: Dict[ScalarConfigItemType, ScalarDefinition] = DEFAULT_SCALAR_DEFINITIONS,
        unknown_scalar_definition: Optional[ScalarDefinition] = None,
        decryption_provider: Optional[BaseDecryptionProvider] = None,
    ):
        """
        Initialize Base Dependencies.

        :param store: The dependency link store to source dependency links from.
        :param scalar_definitions: A map of scalar types from the API definitions to ScalarDefinitions which
        determines how we want map them to concrete Python types and values. Can be overridden to customize
        deserialization behavior or formatting.
        :param unknown_scalar_definition: A scalar definition for handling unknown scalar types from the API. Can be
        used to control behavior for forwards compatibility with new types the SDK does not yet support (e.g.,
        by treating them as strings).
        :param decryption_provider: A decryption provider that can decrypt secrets from app config. If
        dependencies attempt to use a secure_text's decrypted value, a decryption_provider must be specified.
        """
        self._store = store
        self._scalar_definitions = scalar_definitions
        self._unknown_scalar_definition = unknown_scalar_definition
        self._decryption_provider = decryption_provider

    @classmethod
    def from_app(
        cls: Type[D],
        client: Benchling,
        app_id: str,
        decryption_provider: Optional[BaseDecryptionProvider] = None,
    ) -> D:
        """Initialize dependencies from an app_id."""
        link_store = DependencyLinkStore.from_app(client=client, app_id=app_id)
        return cls(link_store, decryption_provider=decryption_provider)

    @classmethod
    def from_store(
        cls: Type[D],
        store: DependencyLinkStore,
        decryption_provider: Optional[BaseDecryptionProvider] = None,
    ) -> D:
        """Initialize dependencies from a store."""
        return cls(store=store, decryption_provider=decryption_provider)

    def invalidate_cache(self) -> None:
        """Invalidate the cache of dependency links and force an update."""
        self._store.invalidate_cache()


def _supported_config_item(config_item: AppConfigItem) -> ConfigurationReference:
    if isinstance(config_item, UnknownType):
        raise UnsupportedDependencyError(
            f"Unable to read app configuration with unsupported type: {config_item}"
        )
    return config_item



from benchling_api_client.v2.beta.api.rna_oligos import bulk_upsert_rna_oligos, upsert_rna_oligo
from benchling_api_client.v2.beta.models.oligo_upsert_request import OligoUpsertRequest
from benchling_api_client.v2.beta.models.rna_oligos_bulk_upsert_request import RnaOligosBulkUpsertRequest

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import AsyncTaskLink, RnaOligo
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaRnaOligoService(BaseService):
    """
    V2-Beta RNA Oligos.

    RNA Oligos are short linear RNA sequences that can be attached as primers to full DNA sequences. Just like other
    entities, they support schemas, tags, and aliases.

    See https://benchling.com/api/v2-beta/reference#/RNA%20Oligos
    """

    @api_method
    def upsert(self, entity_registry_id: str, rna_oligo: OligoUpsertRequest) -> RnaOligo:
        """
        Create or modify a RNA Oligo.

        See https://benchling.com/api/v2-beta/reference#/RNA%20Oligos/upsertRNAOligo
        """
        response = upsert_rna_oligo.sync_detailed(
            client=self.client, entity_registry_id=entity_registry_id, json_body=rna_oligo
        )
        return model_from_detailed(response)

    @api_method
    def bulk_upsert(self, body: RnaOligosBulkUpsertRequest) -> AsyncTaskLink:
        """
        Bulk create or update RNA Oligos.

        See https://benchling.com/api/v2-beta/reference#/RNA%20Oligos/bulkUpsertRnaOligos
        """
        response = bulk_upsert_rna_oligos.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)



from typing import Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.users import (
    bulk_create_users,
    bulk_update_users,
    create_user,
    get_user,
    get_user_activity,
    list_users,
    update_user,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import log_not_implemented
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    AsyncTaskLink,
    ListUsersSort,
    User,
    UserActivity,
    UserBulkCreateRequest,
    UserBulkUpdate,
    UserBulkUpdateRequest,
    UserCreate,
    UsersPaginatedList,
    UserUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class UserService(BaseService):
    """
    Users.

    Benchling users.

    See https://benchling.com/api/reference#/Users
    """

    @api_method
    def get_by_id(self, user_id: str) -> User:
        """
        Get a user by ID.

        See https://benchling.com/api/reference#/Users/getUser
        """
        response = get_user.sync_detailed(client=self.client, user_id=user_id)
        return model_from_detailed(response)

    @api_method
    def _users_page(
        self,
        *,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        member_of: Optional[Iterable[str]] = None,
        admin_of: Optional[Iterable[str]] = None,
        handles: Optional[Iterable[str]] = None,
        password_last_changed_at: Optional[str] = None,
        sort: Optional[ListUsersSort] = None,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
    ) -> Response[UsersPaginatedList]:
        response = list_users.sync_detailed(
            client=self.client,
            ids=none_as_unset(optional_array_query_param(ids)),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            modified_at=none_as_unset(modified_at),
            member_of=none_as_unset(optional_array_query_param(member_of)),
            admin_of=none_as_unset(optional_array_query_param(admin_of)),
            handles=none_as_unset(optional_array_query_param(handles)),
            password_last_changed_at=none_as_unset(password_last_changed_at),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            sort=none_as_unset(sort),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        ids: Optional[Iterable[str]] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        modified_at: Optional[str] = None,
        member_of: Optional[Iterable[str]] = None,
        admin_of: Optional[Iterable[str]] = None,
        handles: Optional[Iterable[str]] = None,
        password_last_changed_at: Optional[str] = None,
        sort: Optional[Union[ListUsersSort]] = None,
        page_size: Optional[int] = 50,
        mentioned_in: Optional[List[str]] = None,
    ) -> PageIterator[User]:
        """
        List users.

        See https://benchling.com/api/reference#/Users/listUsers
        """
        if sort:
            sort = ListUsersSort(sort)

        if mentioned_in:
            log_not_implemented("mentioned_in")

        def api_call(next_token: NextToken) -> Response[UsersPaginatedList]:
            return self._users_page(
                ids=ids,
                name=name,
                name_includes=name_includes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                modified_at=modified_at,
                member_of=member_of,
                admin_of=admin_of,
                handles=handles,
                password_last_changed_at=password_last_changed_at,
                sort=sort,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: UsersPaginatedList) -> Optional[List[User]]:
            return body.users

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, user_create: UserCreate) -> User:
        """
        Create a user.

        See https://benchling.com/api/reference#/Users/createUser
        """
        response = create_user.sync_detailed(client=self.client, json_body=user_create)
        return model_from_detailed(response)

    @api_method
    def update(self, user_id: str, user_update: UserUpdate) -> User:
        """
        Update a user.

        See https://benchling.com/api/reference#/Users/updateUser
        """
        response = update_user.sync_detailed(client=self.client, user_id=user_id, json_body=user_update)
        return model_from_detailed(response)

    @api_method
    def bulk_create(self, user_creates: Iterable[UserCreate]) -> AsyncTaskLink:
        """
        Bulk create users.

        See https://benchling.com/api/reference#/Users/bulkCreateUsers
        """
        body = UserBulkCreateRequest(list(user_creates))
        response = bulk_create_users.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, user_updates: Iterable[UserBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update users.

        See https://benchling.com/api/reference#/Users/bulkUpdateUsers
        """
        body = UserBulkUpdateRequest(list(user_updates))
        response = bulk_update_users.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def get_user_activity(self, user_id: str) -> UserActivity:
        """
        Get activity metadata for a specific user.

        See https://benchling.com/api/reference#/Users/getUserActivity
        """
        response = get_user_activity.sync_detailed(client=self.client, user_id=user_id)
        return model_from_detailed(response)



from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Generic, Iterable, List, Optional, Type, Union

from benchling_api_client.v2.beta.models.canvas_update import CanvasUpdate
from benchling_api_client.v2.beta.models.session import Session
from benchling_api_client.v2.beta.models.session_create import SessionCreate
from benchling_api_client.v2.beta.models.session_message_create import SessionMessageCreate
from benchling_api_client.v2.beta.models.session_message_style import SessionMessageStyle
from benchling_api_client.v2.beta.models.session_status import SessionStatus
from benchling_api_client.v2.beta.models.session_update import SessionUpdate
from benchling_api_client.v2.beta.models.session_update_status import SessionUpdateStatus
from benchling_api_client.v2.stable.types import Unset, UNSET
from typing_extensions import Protocol

from benchling_sdk.apps.framework import AppType
from benchling_sdk.helpers.logging_helpers import log_stability_warning, sdk_logger, StabilityLevel
from benchling_sdk.models import (
    AaSequence,
    Blob,
    Box,
    Container,
    CustomEntity,
    DnaOligo,
    DnaSequence,
    Entry,
    Folder,
    Location,
    Mixture,
    Molecule,
    Plate,
    Request,
    RnaOligo,
    RnaSequence,
    WorkflowOutput,
    WorkflowTask,
)

# Taken from CHIP_SUPPORTED_COLUMN_TYPES in Benchling server
# Anything we miss, they can still embed the ID themselves in a message
_ReferencedSessionLinkType = Union[
    AaSequence,
    Blob,
    Box,
    Container,
    CustomEntity,
    DnaSequence,
    DnaOligo,
    Entry,
    Folder,
    Location,
    Mixture,
    Plate,
    RnaOligo,
    Molecule,
    RnaSequence,
    Request,
    WorkflowOutput,
    WorkflowTask,
]

_DEFAULT_APP_ERROR_MESSAGE = "An unexpected error occurred in the app"

log_stability_warning(StabilityLevel.ALPHA)


class SessionClosedError(Exception):
    """
    Session Closed Error.

    A session was inoperable because its status in Benchling was terminal.
    """

    pass


class SessionContextClosedError(Exception):
    """
    Session Context Closed Error.

    An operation was attempted using the session context manager after it was closed.
    """

    pass


class InvalidSessionTimeoutError(Exception):
    """
    Invalid Session Timeout Error.

    A session's timeout value was set to an invalid value.
    """

    pass


class MissingAttachedCanvasError(Exception):
    """
    Missing Attached Canvas Error.

    A canvas operation was requested, but a session context has no attached canvas.
    """

    pass


class AppUserFacingError(Exception):
    """
    App User Facing Error.

    Extend this class with custom exceptions you want to be written back to the user as a SessionMessage.

    SessionClosingContextExitHandler will invoke messages() and write them to a user. Callers choosing to
    write their own SessionContextExitHandler may need to replicate this behavior themselves.

    This is useful for control flow where an app wants to terminate with an error state that is resolvable
    by the user.

    For example:

    class InvalidUserInputError(AppUserFacingError):
        pass

    raise InvalidUserInputError("Please enter a number between 1 and 10")

    This would create a message shown to the user like:

    SessionMessageCreate("Please enter a number between 1 and 10", style=SessionMessageStyle.ERROR)
    """

    def messages(self) -> List[SessionMessageCreate]:
        """Create a series of SessionMessageCreate to write to a Session and displayed to the user."""
        return [SessionMessageCreate(content=str(self), style=SessionMessageStyle.ERROR)]


class SessionProvider(Protocol):
    """Provide a Session to convey app status."""

    def __call__(self) -> Session:
        """Provide a session when called."""
        pass


class SessionContextErrorProcessor(ABC):
    """
    Session Context Error Processor.

    Implement to match a one or more exception types specified by handles to be processed by process_error
    in the event of a session context exit with errors.

    process_error() does not assume a context's session is active. If using the session with process_error(),
    this should be checked.

    For user with SessionClosingContextExitHandler or its subclasses.
    """

    @classmethod
    @abstractmethod
    def handles(cls) -> List[Type[Exception]]:
        """
        List exception types which should be delegated to this handler.

        Will also match subclasses of that exception type.
        """
        pass

    @classmethod
    @abstractmethod
    def process_error(
        cls,
        context: SessionContextManager[AppType],
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType,
    ) -> bool:
        """
        Process error.

        Receives one of the exception types matched by handles() and takes further action.
        Returns bool to the __exit__ of the context manager.

        If an exception is supplied, and the method wishes to suppress the exception
        (i.e., prevent it from being propagated), it should return a True value.
        Otherwise, the exception will be processed normally upon exit from this method.
        """
        pass


class AppUserFacingErrorProcessor(SessionContextErrorProcessor):
    """
    App User Facing Error Processor.

    An error processor for surfacing error messages directly back to a user.

    Most uncaught exceptions should not be exposed to users to avoid leaking information. App developers
    may curate specific exception types that extend AppUserFacingError. They may stop control flow
    and also write the message directly back to the user.
    """

    @classmethod
    def handles(cls) -> List[Type[Exception]]:
        """Register AppUserFacingError as the types to process."""
        return [AppUserFacingError]

    @classmethod
    def process_error(
        cls,
        context: SessionContextManager[AppType],
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType,
    ) -> bool:
        """
        Process AppUserFacingError.

        Writes a message back to the user with the error's messages and closes the session as FAILED.

        If an exception is supplied, and the method wishes to suppress the exception
        (i.e., prevent it from being propagated), it should return a True value.
        Otherwise, the exception will be processed normally upon exit from this method.
        """
        assert issubclass(
            exc_type, tuple(cls.handles())
        ), f"Expected one of error type {cls.handles()} but got {exc_type}"
        if context.has_active_session():
            assert isinstance(exc_value, AppUserFacingError)
            messages = exc_value.messages()
            context.close_session(status=SessionUpdateStatus.FAILED, messages=messages)
            return False
        else:
            raise SessionClosedError(
                f"Unable to process error of type {exc_type} in exit handler,"
                f" the session is already closed",
                exc_value,
            )


class SessionContextEnterHandler(ABC, Generic[AppType]):
    """
    Session Context Enter Handler.

    An abstract class for defining behavior when a session context's __enter__ method is called.
    """

    @abstractmethod
    def on_enter(self, context: SessionContextManager[AppType]) -> None:
        """Perform on session context enter after a Session has been started with Benchling."""
        pass


class SessionContextExitHandler(ABC, Generic[AppType]):
    """
    Session Context Exit Handler.

    An abstract class for defining behavior when a session context's __exit__ method is called.
    """

    @abstractmethod
    def on_success(self, context: SessionContextManager[AppType]) -> bool:
        """
        Perform on session context exit when no errors are present.

        Returns a bool to indicate context __exit__ behavior.

        If an exception is supplied, and the method wishes to suppress the exception
        (i.e., prevent it from being propagated), it should return a True value.
        Otherwise, the exception will be processed normally upon exit from this method.
        """
        pass

    @abstractmethod
    def on_error(
        self,
        context: SessionContextManager[AppType],
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType,
    ) -> bool:
        """
        Perform when an error caused the session context to exit.

        Returns a bool to indicate context __exit__ behavior.

        If an exception is supplied, and the method wishes to suppress the exception
        (i.e., prevent it from being propagated), it should return a True value.
        Otherwise, the exception will be processed normally upon exit from this method.
        """
        pass


class SessionClosingContextExitHandler(SessionContextExitHandler):
    """
    Session Closing Context Exit Handler.

    Defines behavior when exiting a session context. This implementation only calls the handler

    On success: close the Session with Benchling with SessionStatusUpdate.SUCCEEDED.
    If success_messages were specified, those will also be written to the Session when setting the status.

    On error: close the Session with Benchling with SessionStatusUpdate.FAILED.
    If error_messages were specified, those will also be written to the Session when setting the status.

    As a best practice, error messages are not written to the session to avoid leaking
    potentially sensitive information.
    """

    _success_messages: Optional[Iterable[SessionMessageCreate]]
    _error_messages: Optional[Iterable[SessionMessageCreate]]
    _error_processors: List[Type[SessionContextErrorProcessor]]

    def __init__(
        self,
        success_messages: Optional[Iterable[SessionMessageCreate]] = None,
        error_messages: Optional[Iterable[SessionMessageCreate]] = [
            SessionMessageCreate(_DEFAULT_APP_ERROR_MESSAGE, style=SessionMessageStyle.ERROR)
        ],
        error_processors: List[Type[SessionContextErrorProcessor]] = [AppUserFacingErrorProcessor],
    ):
        """
        Init Session Closing Context Exit Handler.

        Specify an ordered list of SessionContextErrorProcessor types to error_processors for custom
        handling of specific exception types. Processors earlier in the list are matched first.
        """
        self._success_messages = success_messages
        self._error_messages = error_messages
        self._error_processors = error_processors

    def on_success(self, context: SessionContextManager[AppType]) -> bool:
        """
        Close Active Session on Success.

        A SessionSuccessExitHandler which will close an active session when the wrapping session
        context exits.

        Sets the status as SessionUpdateStatus.SUCCEEDED and will set the message to the specified
        text, if provided, with a SessionMessageStyle.SUCCESS.

        Generally prefer apps to manually close_session() with more informative success
        messages, but this provides a fallback in case the context exits without error first.
        """
        active_session = context.active_session()
        if active_session:
            sdk_logger.debug(
                "Exiting session context, automatically closing session ID %s as %s",
                active_session.id,
                SessionUpdateStatus.SUCCEEDED,
            )
            if self._success_messages:
                message_creates = _ordered_messages(self._success_messages)
                context.close_session(SessionUpdateStatus.SUCCEEDED, messages=message_creates)
            else:
                context.close_session(SessionUpdateStatus.SUCCEEDED)
        else:
            sdk_logger.debug("Exiting session context, no active session present")
        return True

    def on_error(
        self,
        context: SessionContextManager[AppType],
        exc_type: Type[BaseException],
        exc_value: BaseException,
        exc_traceback: TracebackType,
    ) -> bool:
        """
        Close Active Session on Error.

        A SessionErrorExitHandler which will close an active session when the wrapping session
        context exits with any type of Exception.

        Sets the status as SessionUpdateStatus.FAILED and will set the message to the specified
        text, if provided, with a SessionMessageStyle.ERROR.

        If an exception is supplied, and the method wishes to suppress the exception
        (i.e., prevent it from being propagated), it should return a True value.
        Otherwise, the exception will be processed normally upon exit from this method.
        """
        active_session = context.active_session()
        if active_session:
            optional_error_processor = self._error_processor_type(exc_type)
            # Pluggable specific error handling
            if optional_error_processor:
                sdk_logger.debug("Exiting session context with error, matched handler for %s", str(exc_type))
                return optional_error_processor.process_error(context, exc_type, exc_value, exc_traceback)
            # Fallback to general error handling
            sdk_logger.debug(
                "Exiting session context with error, automatically closing session ID %s as %s",
                active_session.id,
                SessionUpdateStatus.FAILED,
            )
            sdk_logger.debug(exc_value)
            sdk_logger.debug(exc_traceback)
            if self._error_messages:
                message_creates = _ordered_messages(self._error_messages)
                context.close_session(SessionUpdateStatus.FAILED, messages=message_creates)
            else:
                context.close_session(SessionUpdateStatus.FAILED)
        else:
            sdk_logger.debug("Exiting session context, no active session present")
        return False

    def _error_processor_type(
        self, exc_type: Type[BaseException]
    ) -> Optional[Type[SessionContextErrorProcessor]]:
        for processor in self._error_processors:
            for error_type in processor.handles():
                if issubclass(exc_type, error_type):
                    return processor
        return None


def create_session_provider(app: AppType, name: str, timeout_seconds: int) -> SessionProvider:
    """
    Create Session Provider.

    Create a SessionProvider that will create a new session with the specified app. This is the
    preferred way of working with sessions for most apps.
    """

    def _new_session() -> Session:
        session_create = SessionCreate(app_id=app.id, name=name, timeout_seconds=timeout_seconds)
        return app.benchling.v2.beta.apps.create_session(session_create)

    return _new_session


def existing_session_provider(app: AppType, session_id: str) -> SessionProvider:
    """
    Existing Session Provider.

    Create a SessionProvider that will fetch the existing session by ID using the app's Benchling instance.
    This might be used in cases like distributed apps which may be sharing a session and passing a
    reference to the handle (session ID).
    """

    def _existing_session() -> Session:
        session = app.benchling.v2.beta.apps.get_session_by_id(session_id)
        if session.status != SessionStatus.RUNNING:
            raise SessionClosedError(
                f"Cannot continue session with ID {session_id} with status {session.status}"
            )
        return session

    return _existing_session


def _ordered_messages(messages: Iterable[SessionMessageCreate]) -> List[SessionMessageCreate]:
    """Coerce messages into an ordered list."""
    return list(messages)


class SessionContextManager(AbstractContextManager, Generic[AppType]):
    """
    Manage Benchling App Session.

    On init, will invoke SessionProvider to get a session accordingly. When exiting the session context,
    will invoke error_exit_handler if an exception is encountered. Otherwise, invokes the
    success_exit_handler.
    """

    _app: AppType
    _session_provider: SessionProvider
    _context_enter_handler: Optional[SessionContextEnterHandler[AppType]]
    _context_exit_handler: SessionContextExitHandler[AppType]
    _session: Optional[Session]
    _attached_canvas_id: Optional[str]

    def __init__(
        self,
        app: AppType,
        session_provider: SessionProvider,
        context_enter_handler: Optional[SessionContextEnterHandler[AppType]] = None,
        context_exit_handler: Optional[SessionContextExitHandler[AppType]] = None,
    ):
        """
        Initialize SessionContextManager.

        Prefer new_session_context() or continue_session_context() over initializing this directly
        unless it's necessary to extend this class.
        """
        self._app = app
        self._session_provider = session_provider
        self._context_enter_handler = context_enter_handler
        self._context_exit_handler = (
            context_exit_handler if context_exit_handler is not None else SessionClosingContextExitHandler()
        )
        self._session = None
        self._attached_canvas_id = None

    def __enter__(self):
        self._session = self._session_provider()
        if self._context_enter_handler:
            self._context_enter_handler.on_enter(self)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> bool:
        if exc_type:
            # Appease MyPy
            assert exc_value is not None
            assert exc_traceback is not None
            return self._context_exit_handler.on_error(self, exc_type, exc_value, exc_traceback)
        return self._context_exit_handler.on_success(self)

    @property
    def app(self) -> AppType:
        """Return the app for the session."""
        return self._app

    def write_strings(
        self, contents: Union[str, Iterable[str]], style: Union[SessionMessageStyle, Unset] = UNSET
    ) -> None:
        """
        Write Session Messages from string contents.

        Shorthand for writing session messages without creating a SessionMessageCreate instance.

        Write one or more messages to app's session from the specified contents string(s). If passed an iterable,
        each element will create its own session message. If style is specified, it will be applied to all
        messages.

        For more granular control, use write_messages().

        Requires an active session which is unclosed.
        """
        if isinstance(contents, str):
            messages_to_create = [SessionMessageCreate(contents, style=style)]
        else:
            messages_to_create = [SessionMessageCreate(text, style=style) for text in contents]
        session_update = SessionUpdate(messages=messages_to_create)
        self._update_session(session_update)

    def write_messages(self, messages: Iterable[SessionMessageCreate]) -> None:
        """
        Write Session Messages.

        Write one or more messages to app's session. Requires an active session which is unclosed.
        """
        session_update = SessionUpdate(messages=_ordered_messages(messages))
        self._update_session(session_update)

    def add_timeout_seconds(self, seconds_to_add: int) -> None:
        """
        Add Session Timeout Seconds.

        Add seconds to the session's timeout in Benchling. Requires an active session which is unclosed.
        """
        if seconds_to_add < 1:
            raise InvalidSessionTimeoutError("The session timeout must be a positive integer in seconds.")
        self._require_session()
        # Appease MyPy
        assert self._session is not None
        new_timeout = self._session.timeout_seconds + seconds_to_add
        session_update = SessionUpdate(timeout_seconds=new_timeout)
        self._update_session(session_update)

    def set_timeout_seconds(self, timeout_seconds: int) -> None:
        """
        Set Session Timeout.

        Sets the session's timeout in Benchling. Requires an active session which is unclosed.

        Timeouts may only be increased, up to the Benchling allowed maximum.
        """
        self._require_session()
        # Appease MyPy
        assert self._session is not None
        if timeout_seconds < self._session.timeout_seconds:
            raise InvalidSessionTimeoutError(
                f"The session timeout must be a positive integer in seconds larger than the "
                f"existing timeout ({self._session.timeout_seconds} seconds)"
            )
        session_update = SessionUpdate(timeout_seconds=timeout_seconds)
        self._update_session(session_update)

    def close_session(
        self, status: SessionUpdateStatus, messages: Optional[Iterable[SessionMessageCreate]] = None
    ):
        """
        Close Session.

        Closes the session with Benchling on the server. This MUST be the last call in the session
        context.

        This will close the session. Subsequent attempts to perform session operations
        may raise SessionContextClosedError.

        Do not continue performing operations in the session context after calling close_session().
        """
        self._require_session()
        messages_or_unset: Union[List[SessionMessageCreate], Unset] = (
            _ordered_messages(messages) if messages is not None else UNSET
        )
        session_update = SessionUpdate(status=status, messages=messages_or_unset)
        self._update_session(session_update)
        self._session = None

    def _update_session(self, session_update: SessionUpdate) -> None:
        self._require_session()
        active_session = self.active_session()
        # Appease MyPy, which isn't aware of how _require_session() asserts this isn't none
        assert active_session is not None
        self._session = self._app.benchling.v2.beta.apps.update_session(active_session.id, session_update)

    def active_session(self) -> Optional[Session]:
        """
        Active Session.

        Return an active Session held by the session context. Return None if the context does not have
        an active Session.
        """
        return self._session

    def has_active_session(self) -> bool:
        """
        Has Active Session.

        Return True if the active Session with Benchling has not been closed by the session context.
        """
        return self._session is not None

    def attach_canvas(self, canvas_id: str) -> None:
        """
        Attach Canvas.

        Updates the specified canvas_id with the active session_id,
        causing session information to display on the specified canvas.
        """
        self._require_session()
        active_session = self.active_session()
        # Appease MyPy, which isn't aware of how _require_session() asserts this isn't none
        assert active_session is not None
        # TODO BNCH-57234 we can't update canvas without setting blocks and featureId
        canvas = self.app.benchling.v2.beta.apps.get_canvas(canvas_id)
        canvas_update = CanvasUpdate(
            session_id=active_session.id, blocks=canvas.blocks, feature_id=canvas.feature_id
        )
        updated_canvas = self.app.benchling.v2.beta.apps.update_canvas(canvas_id, canvas_update)
        sdk_logger.debug(
            "Attached session context with session id %s to canvas id %s",
            active_session.id,
            updated_canvas.id,
        )
        self._attached_canvas_id = updated_canvas.id

    def detach_canvas(self) -> None:
        """
        Detach Canvas.

        Updates the attached canvas in the context to remove the session_id. This will cause the canvas to cease
        displaying session information and messages.

        Requires that attach_canvas() was called previously, or an error will be raised.
        """
        if self._attached_canvas_id:
            # TODO BNCH-57234 we can't update canvas without setting blocks and featureId
            canvas = self.app.benchling.v2.beta.apps.get_canvas(self._attached_canvas_id)
            canvas_update = CanvasUpdate(session_id=None, blocks=canvas.blocks, feature_id=canvas.feature_id)
            updated_canvas = self.app.benchling.v2.beta.apps.update_canvas(
                self._attached_canvas_id, canvas_update
            )
            self._attached_canvas_id = None
            sdk_logger.debug("Detached session context from canvas id %s", updated_canvas.id)
        else:
            raise MissingAttachedCanvasError("The session context does not have an attached canvas to detach")

    def _require_session(self):
        """Raise SessionContextClosedError if a context operation requires an active session."""
        if not self.has_active_session():
            raise SessionContextClosedError("The app session in the context has already been closed")


def new_session_context(
    app: AppType,
    name: str,
    timeout_seconds: int,
    context_enter_handler: Optional[SessionContextEnterHandler] = None,
    context_exit_handler: Optional[SessionContextExitHandler] = None,
) -> SessionContextManager[AppType]:
    """
    Create New Session Context.

    Create a context manager that will provision a new app Session with Benchling.
    """
    return SessionContextManager(
        app,
        create_session_provider(app, name, timeout_seconds),
        context_enter_handler,
        context_exit_handler,
    )


def continue_session_context(
    app: AppType,
    session_id: str,
    context_enter_handler: Optional[SessionContextEnterHandler] = None,
    context_exit_handler: Optional[SessionContextExitHandler] = None,
) -> SessionContextManager[AppType]:
    """
    Continue Session Context.

    Create a context manager that will fetch an existing Session from Benchling and create a context with it.
    """
    return SessionContextManager(
        app,
        existing_session_provider(app, session_id),
        context_enter_handler,
        context_exit_handler,
    )


def ref(reference: _ReferencedSessionLinkType) -> str:
    """
    Ref.

    Helper method for easily serializing a referenced object into a string embeddable in SessionMessageCreate
    content.

    Example:
        dna_sequence = benchling.dna_sequences.get_by_id("seq_1234")
        SessionMessageCreate(f"This is my DNA sequence {ref(dna_sequence)} for analysis"
    """
    # Not sure {} are possible in Benchling IDs, but the spec says we're escaping
    unescape_id = reference.id
    escaped_id = unescape_id.replace("{", "\\{").replace("}", "\\}")
    return f"{{id:{escaped_id}}}"



from __future__ import annotations

from typing import cast, Dict, Generic, List, Optional, Type, TypeVar, Union

from benchling_api_client.v2.beta.models.button_ui_block import ButtonUiBlock
from benchling_api_client.v2.beta.models.canvas import Canvas
from benchling_api_client.v2.beta.models.canvas_create import CanvasCreate
from benchling_api_client.v2.beta.models.canvas_update import CanvasUpdate
from benchling_api_client.v2.beta.models.chip_ui_block import ChipUiBlock
from benchling_api_client.v2.beta.models.dropdown_ui_block import DropdownUiBlock
from benchling_api_client.v2.beta.models.markdown_ui_block import MarkdownUiBlock
from benchling_api_client.v2.beta.models.search_input_ui_block import SearchInputUiBlock
from benchling_api_client.v2.beta.models.section_ui_block import SectionUiBlock
from benchling_api_client.v2.beta.models.selector_input_ui_block import SelectorInputUiBlock
from benchling_api_client.v2.beta.models.text_input_ui_block import TextInputUiBlock
from benchling_api_client.v2.extensions import UnknownType
from typing_extensions import Protocol

from benchling_sdk.helpers.logging_helpers import log_stability_warning, StabilityLevel

log_stability_warning(StabilityLevel.ALPHA)

S = TypeVar("S", bound="_FilteredCanvasBuilderBlockStream")

_UiBlock = Union[
    ButtonUiBlock,
    ChipUiBlock,
    DropdownUiBlock,
    MarkdownUiBlock,
    SearchInputUiBlock,
    SectionUiBlock,
    SelectorInputUiBlock,
    TextInputUiBlock,
    UnknownType,
]

_UiBlockType = TypeVar(
    "_UiBlockType",
    bound=_UiBlock,
)


class DuplicateBlockIdError(Exception):
    """Error indicating that duplicate ids were present on blocks within a Canvas."""

    pass


class NoMatchingBlocksError(Exception):
    """
    Error indicating that blocks were expected, but none matched.

    Used to prevent requiring developers to handle Optional[_UiBlock] for type safety.
    """

    pass


class _CanvasBuilderUiBlock(Generic[_UiBlockType]):
    """Internal UI block wrapper for CanvasBuilder."""

    _block: _UiBlockType
    _builder: CanvasBuilder

    def __init__(self, block: _UiBlockType, builder: CanvasBuilder):
        self._block = block
        self._builder = builder

    @classmethod
    def from_api_model(
        cls, block: _UiBlockType, builder: CanvasBuilder
    ) -> _CanvasBuilderUiBlock[_UiBlockType]:
        """Create a _CanvasBuilderUiBlock from an underlying API model."""
        return cls(block, builder)

    def to_api_model(self) -> _UiBlockType:
        """Convert to the underlying API model."""
        return self._block

    def children(self) -> _CanvasBuilderBlockStream:
        """
        Return children for blocks when applicable, such as for section blocks.

        If not applicable, returns an empty canvas block stream with no blocks.
        """
        model = self.to_api_model()
        if isinstance(model, SectionUiBlock):
            # MyPy can't recognize the type narrowing when we check .children below
            section_block = cast(SectionUiBlock, model)
            child_blocks = [
                _CanvasBuilderUiBlock.from_api_model(block, self._builder) for block in section_block.children
            ]
            # Pass reference to parent block (self)
            return _CanvasBuilderBlockStream(self._builder, child_blocks, child_blocks, self)
        return _CanvasBuilderBlockStream(self._builder, [], [])

    def replace(self, new_blocks: List[_UiBlock]) -> None:
        """Replace block with provided new_blocks."""
        parent = self._parent_block()
        model = cast(_UiBlock, self.to_api_model())
        if parent:
            self.insert_after(new_blocks)
            # Keeps MyPy happy; SectionUiBlock is not valid for .children of SectionUiBlock
            assert not isinstance(model, SectionUiBlock)
            parent.children.remove(model)
        else:
            self.insert_after(new_blocks)
            # noinspection PyProtectedMember
            self._builder._source_canvas.blocks.remove(model)

    def remove(self) -> None:
        """Remove block."""
        parent = self._parent_block()
        model = cast(_UiBlock, self.to_api_model())
        if parent:
            # Keeps MyPy happy; SectionUiBlock is not valid for .children of SectionUiBlock
            assert not isinstance(model, SectionUiBlock)
            parent.children.remove(model)
        else:
            # noinspection PyProtectedMember
            self._builder._source_canvas.blocks.remove(model)

    def insert_after(self, new_blocks: List[_UiBlock]) -> None:
        """Insert new_blocks after block."""
        self._nested_insert(new_blocks, 1)

    def insert_before(self, new_blocks: List[_UiBlock]) -> None:
        """Insert new_blocks before block."""
        self._nested_insert(new_blocks, 0)

    def _nested_insert(self, new_blocks: List[_UiBlock], offset: int) -> None:
        """
        Nested insert.

        This is used to handle the case that an insert is being performed and it's contextual within
        another block.

        For instance, the children of a SectionUiBlock.
        """
        parent_block = self._parent_block()
        if parent_block:
            child_blocks = parent_block.children
            # Using list() to solve "List" is invariant creates a copy which means this stops working
            self._insert(new_blocks, offset, child_blocks, self.to_api_model())  # type: ignore
            parent_block.children = child_blocks
            parent_builder_block = _CanvasBuilderUiBlock.from_api_model(parent_block, self._builder)
            parent_builder_block.insert_after([parent_block])
            # noinspection PyProtectedMember
            self._builder._source_canvas.blocks.remove(parent_builder_block.to_api_model())
        else:
            # noinspection PyProtectedMember
            self._insert(new_blocks, offset, self._builder._source_canvas.blocks, self.to_api_model())

    @staticmethod
    def _insert(
        new_blocks: List[_UiBlock], offset: int, blocks: List[_UiBlock], target_block: _UiBlock
    ) -> None:
        """Insert new_blocks before block as a side effect."""
        index = blocks.index(target_block)
        for count, new_block in enumerate(new_blocks):
            blocks.insert(index + count + offset, new_block)

    def _parent_block(self) -> Optional[SectionUiBlock]:
        # noinspection PyProtectedMember
        for parent_block in self._builder._source_canvas.blocks:
            if isinstance(parent_block, SectionUiBlock):
                if self.to_api_model() in parent_block.children:
                    return parent_block
        return None


class _CanvasBuilderFilter(Protocol):
    def __call__(self, block: _UiBlockType) -> bool:
        """Return True if the UiBlock matches specified conditions."""
        pass


class _FilteredCanvasBuilderBlockStream:
    """Internal Filtered UI block list wrapper for CanvasBuilder."""

    _builder: CanvasBuilder
    _blocks: List[_CanvasBuilderUiBlock]
    _selected_blocks: List[_CanvasBuilderUiBlock]
    _cursor: int

    def __init__(
        self,
        builder: CanvasBuilder,
        blocks: List[_CanvasBuilderUiBlock],
        selected_blocks: List[_CanvasBuilderUiBlock],
    ):
        self._builder = builder
        self._blocks = blocks
        self._selected_blocks = selected_blocks

    def __iter__(self):
        self._cursor = 0
        return self

    def __next__(self):
        if self._cursor >= len(self._blocks):
            raise StopIteration
        block = self._blocks[self._cursor]
        self._cursor += 1
        return block

    @classmethod
    def from_builder(cls: Type[S], builder: CanvasBuilder) -> S:
        """
        From Builder.

        Instantiate a new _FilteredCanvasBuilderBlockStream from a CanvasBuilder.
        """
        # noinspection PyProtectedMember
        blocks = [
            _CanvasBuilderUiBlock.from_api_model(block, builder)
            for block in builder._source_canvas.blocks
            if not isinstance(block, UnknownType)
        ]
        return cls(builder, blocks, blocks)

    def filter(self, filter_function: _CanvasBuilderFilter) -> _FilteredCanvasBuilderBlockStream:
        """
        Filter.

        Accept a predicate that evaluates if a UiBlock should be included in the result or not.
        Returns a new stream of blocks filtered to the predicate, which is further operable.
        """
        return _CanvasBuilderBlockStream(
            self._builder,
            self._blocks,
            [block for block in self._selected_blocks if filter_function(block.to_api_model())],
        )

    def get_by_id(self, block_id: str) -> _CanvasBuilderUiBlock:
        """
        Get a block by its id.

        Raises NoMatchingBlocksError if the block is not found. To match an id as an optional, use filter().
        """
        matched_block = self._block_by_id(block_id, self._blocks)
        if not matched_block:
            raise NoMatchingBlocksError(f'Could not find a block with id "{block_id}"')
        return matched_block

    def _block_by_id(
        self, block_id: str, blocks: List[_CanvasBuilderUiBlock]
    ) -> Optional[_CanvasBuilderUiBlock]:
        for block in blocks:
            api_block = block.to_api_model()
            if api_block.id == block_id:
                return block
            children = block.children()
            if children.count() > 0:
                child_block = self._block_by_id(block_id, list(children))
                if child_block:
                    return child_block
        return None

    def count(self) -> int:
        """Return a count of the elements in the list of blocks."""
        return len(self._selected_blocks)

    def first(self) -> _CanvasBuilderUiBlock:
        """Return the first block in the list."""
        if len(self._selected_blocks) < 1:
            raise NoMatchingBlocksError
        return self._selected_blocks[0]

    def last(self) -> _CanvasBuilderUiBlock:
        """Return the last block in the list."""
        if len(self._selected_blocks) < 1:
            raise NoMatchingBlocksError
        return self._selected_blocks[-1]

    def remove(self) -> None:
        """Remove blocks."""
        updated_blocks = [
            block.to_api_model() for block in self._blocks if block not in self._selected_blocks
        ]
        # noinspection PyProtectedMember
        self._builder._source_canvas.blocks = updated_blocks


class _CanvasBuilderBlockStream(_FilteredCanvasBuilderBlockStream):
    """
    Internal UI block list wrapper for CanvasBuilder.

    Possesses some additional operations unavailable to filtered block streams.
    """

    _parent: Optional[_CanvasBuilderUiBlock]

    def __init__(
        self,
        builder: CanvasBuilder,
        blocks: List[_CanvasBuilderUiBlock],
        selected_blocks: List[_CanvasBuilderUiBlock],
        parent: Optional[_CanvasBuilderUiBlock] = None,
    ):
        super().__init__(builder, blocks, selected_blocks)
        self._parent = parent

    def append(self, new_blocks: List[_UiBlock]) -> None:
        """
        Append new_blocks to the end of list of blocks.

        Only operates on unfiltered block streams. This can be a list of blocks on a Canvas, or children on a
        block such as SectionUiBlock.
        """
        if self._parent:
            if self.count() > 0:
                self.last().insert_after(new_blocks)
            else:
                self._parent.to_api_model().children.extend(new_blocks)
        else:
            # noinspection PyProtectedMember
            self._builder._source_canvas.blocks.extend(new_blocks)


class CanvasBuilder:
    """
    Canvas Builder.

    This class provides methods to help developers effectively work with Canvas UI Blocks.
    Working with the underlying API models directly can be clunky, as there is no native way to easily find
    blocks based on their attributes, and then operate on them. Some blocks, like SectionUiBlock, have nested
    children blocks, further complicating the unassisted experience.

    The goal of CanvasBuilder is to accept an existing Canvas, and easily change its contents (blocks), then
    send a resulting CanvasCreate or CanvasUpdate model to the API.

    Some block operations include:
        - Get by Id
        - Filtering by predicate
        - Remove
        - Replace
        - Insert After
        - Insert Before
        - Append

    Sample usage:
        ```
        canvas: Canvas = benchling.v2.beta.apps.get_canvas_by_id("canvas_id")
        builder = CanvasBuilder.from_canvas(canvas).blocks
            .filter(lambda block: isinstance(block, TextInputUiBlock))
            .remove()
        updated_canvas = benchling.v2.beta.apps.update_canvas("canvas_id", builder.to_update())
        ```
    """

    _source_canvas: Canvas

    def __init__(
        self,
        feature_id: str,
        resource_id: str,
        enabled: bool = True,
        session_id: Optional[str] = None,
        blocks: Optional[List[_UiBlock]] = None,
    ):
        """
        Init Canvas.

        Create a CanvasBuilder from scratch. Useful when a source Canvas is not already created.
        """
        self._source_canvas = Canvas(
            feature_id=feature_id,
            resource_id=resource_id,
            enabled=enabled,
            session_id=session_id,
            blocks=blocks if blocks else [],
        )

    @classmethod
    def from_canvas(cls, canvas: Canvas) -> CanvasBuilder:
        """
        From Canvas.

        Create a CanvasBuilder from an existing canvas. Preferred when a canvas already exists.
        """
        return cls(
            feature_id=canvas.feature_id,
            resource_id=canvas.resource_id,
            enabled=canvas.enabled,
            session_id=canvas.session_id,
            blocks=canvas.blocks,
        )

    def _set_enabled(self, value: bool) -> CanvasBuilder:
        return CanvasBuilder(
            feature_id=self._source_canvas.feature_id,
            resource_id=self._source_canvas.resource_id,
            enabled=value,
            session_id=self._source_canvas.session_id,
            blocks=self._source_canvas.blocks,
        )

    def disable(self) -> CanvasBuilder:
        """Set the canvas to disabled."""
        return self._set_enabled(False)

    def enable(self) -> CanvasBuilder:
        """Set the canvas to enabled."""
        return self._set_enabled(True)

    def inputs_to_dict(self) -> Dict[str, str]:
        """
        Read Inputs to dict.

        Return a dictionary of {block_id: block_value} for all blocks on the canvas with input values.
        Blocks that only have read attributes are omitted.

        Raise DuplicateBlockIdError if multiple blocks with the same id are found.
        """
        return self._values_from_blocks(self._source_canvas.blocks)

    def _values_from_blocks(
        self, blocks: List[_UiBlock], existing_keys: Optional[List[str]] = None
    ) -> Dict[str, str]:
        existing_keys = existing_keys if existing_keys else []
        values: Dict[str, str] = dict()
        for block in blocks:
            if not isinstance(block, UnknownType):
                if hasattr(block, "value"):
                    if block.id in existing_keys:
                        raise DuplicateBlockIdError(
                            f'More than one block with the id "{block.id}" already exists'
                        )
                    existing_keys.append(block.id)
                    # Ignore type error since we checked hasattr above
                    values[block.id] = block.value  # type: ignore
                elif isinstance(block, SectionUiBlock):
                    # Cast to appease MyPy
                    parent_block = cast(SectionUiBlock, block)
                    # list() prevents MyPy from complaining: "List" is invariant
                    child_values = self._values_from_blocks(list(parent_block.children), existing_keys)
                    values.update(child_values)
        return values

    def to_update(self) -> CanvasUpdate:
        """Return a CanvasUpdate API model from the current state of the canvas managed by CanvasBuilder."""
        return CanvasUpdate(
            feature_id=self._source_canvas.feature_id,
            resource_id=self._source_canvas.resource_id,
            enabled=self._source_canvas.enabled,
            session_id=self._source_canvas.session_id,
            blocks=self._source_canvas.blocks,
        )

    def to_create(self) -> CanvasCreate:
        """Return a CanvasCreate API model from the current state of the canvas managed by CanvasBuilder."""
        return CanvasCreate(
            feature_id=self._source_canvas.feature_id,
            resource_id=self._source_canvas.resource_id,
            enabled=self._source_canvas.enabled,
            session_id=self._source_canvas.session_id,
            blocks=self._source_canvas.blocks,
        )

    @property
    def blocks(self) -> _CanvasBuilderBlockStream:
        """
        Blocks.

        Return a stream of blocks which can be iterated and operated on to mutate the canvas
        stored by the builder.
        """
        return _CanvasBuilderBlockStream.from_builder(self)



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.dna_sequences import (
    archive_dna_sequences,
    auto_annotate_dna_sequences,
    autofill_dna_sequence_parts,
    autofill_dna_sequence_translations,
    bulk_create_dna_sequences,
    bulk_get_dna_sequences,
    bulk_update_dna_sequences,
    create_dna_sequence,
    get_dna_sequence,
    list_dna_sequences,
    unarchive_dna_sequences,
    update_dna_sequence,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AsyncTaskLink,
    AutoAnnotateDnaSequences,
    AutofillSequences,
    DnaSequence,
    DnaSequenceBulkCreate,
    DnaSequenceBulkUpdate,
    DnaSequenceCreate,
    DnaSequencesArchivalChange,
    DnaSequencesArchive,
    DnaSequencesBulkCreateRequest,
    DnaSequencesBulkUpdateRequest,
    DnaSequencesPaginatedList,
    DnaSequencesUnarchive,
    DnaSequenceUpdate,
    EntityArchiveReason,
    ListDNASequencesSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class DnaSequenceService(BaseService):
    """
    DNA Sequences.

    DNA sequences are the bread and butter of the Benchling Molecular Biology suite. On Benchling, these are
    comprised of a string of nucleotides and collections of other attributes, such as annotations and primers.

    See https://benchling.com/api/reference#/DNA%20Sequences
    """

    @api_method
    def get_by_id(self, dna_sequence_id: str) -> DnaSequence:
        """
        Get a DNA sequence.

        See https://benchling.com/api/reference#/DNA%20Sequences/getDNASequence
        """
        response = get_dna_sequence.sync_detailed(client=self.client, dna_sequence_id=dna_sequence_id)
        return model_from_detailed(response)

    @api_method
    def _dna_sequences_page(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        bases: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        sort: Optional[ListDNASequencesSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        next_token: NextToken = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> Response[DnaSequencesPaginatedList]:
        response = list_dna_sequences.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            bases=none_as_unset(bases),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            schema_id=none_as_unset(schema_id),
            archive_reason=none_as_unset(archive_reason),
            mentions=none_as_unset(optional_array_query_param(mentions)),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        bases: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        sort: Optional[Union[str, ListDNASequencesSort]] = None,
        page_size: Optional[int] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> PageIterator[DnaSequence]:
        """
        List DNA sequences.

        See https://benchling.com/api/reference#/DNA%20Sequences/listDNASequences
        """
        check_for_csv_bug_fix("mentioned_in", mentioned_in)
        check_for_csv_bug_fix("mentions", mentions)

        def api_call(next_token: NextToken) -> Response[DnaSequencesPaginatedList]:
            return self._dna_sequences_page(
                modified_at=modified_at,
                name=name,
                bases=bases,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                schema_id=schema_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                name_includes=name_includes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                schema_fields=schema_fields,
                sort=_translate_to_string_enum(ListDNASequencesSort, sort),
                page_size=page_size,
                next_token=next_token,
                author_idsany_of=author_idsany_of,
            )

        def results_extractor(body: DnaSequencesPaginatedList) -> Optional[List[DnaSequence]]:
            return body.dna_sequences

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, dna_sequence: DnaSequenceCreate) -> DnaSequence:
        """
        Create a DNA sequence.

        See https://benchling.com/api/reference#/DNA%20Sequences/createDNASequence
        """
        response = create_dna_sequence.sync_detailed(client=self.client, json_body=dna_sequence)
        return model_from_detailed(response)

    @api_method
    def update(self, dna_sequence_id: str, dna_sequence: DnaSequenceUpdate) -> DnaSequence:
        """
        Update a DNA sequence.

        See https://benchling.com/api/reference#/DNA%20Sequences/updateDNASequence
        """
        response = update_dna_sequence.sync_detailed(
            client=self.client, dna_sequence_id=dna_sequence_id, json_body=dna_sequence
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, dna_sequence_ids: Iterable[str], reason: EntityArchiveReason
    ) -> DnaSequencesArchivalChange:
        """
        Archive DNA sequences.

        See https://benchling.com/api/reference#/DNA%20Sequences/archiveDNASequences
        """
        archive_request = DnaSequencesArchive(reason=reason, dna_sequence_ids=list(dna_sequence_ids))
        response = archive_dna_sequences.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, dna_sequence_ids: Iterable[str]) -> DnaSequencesArchivalChange:
        """
        Unarchive DNA sequences.

        See https://benchling.com/api/reference#/DNA%20Sequences/unarchiveDNASequences
        """
        unarchive_request = DnaSequencesUnarchive(dna_sequence_ids=list(dna_sequence_ids))
        response = unarchive_dna_sequences.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_get(self, dna_sequence_ids: Iterable[str]) -> Optional[List[DnaSequence]]:
        """
        Bulk get DNA sequences.

        See https://benchling.com/api/reference#/DNA%20Sequences/bulkGetDNASequences
        """
        dna_sequence_id_string = ",".join(dna_sequence_ids)
        response = bulk_get_dna_sequences.sync_detailed(
            client=self.client, dna_sequence_ids=dna_sequence_id_string
        )
        dna_sequences_results = model_from_detailed(response)
        return dna_sequences_results.dna_sequences

    @api_method
    def bulk_create(self, dna_sequences: Iterable[DnaSequenceBulkCreate]) -> AsyncTaskLink:
        """
        Bulk create DNA sequences.

        See https://benchling.com/api/reference#/DNA%20Sequences/bulkCreateDNASequences
        """
        body = DnaSequencesBulkCreateRequest(list(dna_sequences))
        response = bulk_create_dna_sequences.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, dna_sequences: Iterable[DnaSequenceBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update DNA sequences.

        See https://benchling.com/api/reference#/DNA%20Sequences/bulkUpdateDNASequences
        """
        body = DnaSequencesBulkUpdateRequest(list(dna_sequences))
        response = bulk_update_dna_sequences.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def autofill_parts(self, dna_sequence_ids: Iterable[str]) -> AsyncTaskLink:
        """
        Autofill DNA sequence parts.

        See https://benchling.com/api/reference#/DNA%20Sequences/autofillDNASequenceParts
        """
        body = AutofillSequences(dna_sequence_ids=list(dna_sequence_ids))
        response = autofill_dna_sequence_parts.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def autofill_translations(self, dna_sequence_ids: Iterable[str]) -> AsyncTaskLink:
        """
        Autofill DNA sequence translations.

        See https://benchling.com/api/reference#/DNA%20Sequences/autofillDNASequenceTranslations
        """
        body = AutofillSequences(dna_sequence_ids=list(dna_sequence_ids))
        response = autofill_dna_sequence_translations.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def auto_annotate(self, auto_annotate: AutoAnnotateDnaSequences) -> AsyncTaskLink:
        """
        Auto-annotate DNA sequences with matching features from specified Feature Libraries.

        See https://benchling.com/api/reference#/DNA%20Sequences/autoAnnotateDnaSequences
        """
        response = auto_annotate_dna_sequences.sync_detailed(client=self.client, json_body=auto_annotate)
        return model_from_detailed(response)



import os
from pathlib import Path

docs_path = "file://" + str(Path(Path(os.path.realpath(__file__)).parent, "html", "index.html").absolute())
try:
    import webbrowser

    webbrowser.open(docs_path)
except ImportError:
    print(docs_path)



from benchling_api_client.v2.beta.api.aa_sequences import (
    bulk_upsert_aa_sequences,
    find_matching_regions_aa_sequences,
    upsert_aa_sequence,
)
from benchling_api_client.v2.beta.models.aa_sequence_upsert import AaSequenceUpsert
from benchling_api_client.v2.beta.models.aa_sequences_bulk_upsert_request import AaSequencesBulkUpsertRequest
from benchling_api_client.v2.beta.models.aa_sequences_find_matching_region import (
    AaSequencesFindMatchingRegion,
)

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import AaSequence, AsyncTaskLink
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaAaSequenceService(BaseService):
    """
    V2-Beta AA Sequences.

    AA Sequences are the working units of cells that make everything run (they help make structures, catalyze
    reactions and allow for signaling - a kind of internal cell communication). On Benchling, these are comprised
    of a string of amino acids and collections of other attributes, such as annotations.

    See https://benchling.com/api/v2-beta/reference#/AA%20Sequences
    """

    @api_method
    def upsert(self, entity_registry_id: str, aa_sequence: AaSequenceUpsert) -> AaSequence:
        """
        Create or modify an Amino Acid Sequence.

        See https://benchling.com/api/v2-beta/reference#/AA%20Sequences/upsertAaSequence
        """
        response = upsert_aa_sequence.sync_detailed(
            client=self.client, entity_registry_id=entity_registry_id, json_body=aa_sequence
        )
        return model_from_detailed(response)

    @api_method
    def bulk_upsert(self, body: AaSequencesBulkUpsertRequest) -> AsyncTaskLink:
        """
        Bulk create or update AA sequences.

        See https://benchling.com/api/v2-beta/reference#/AA%20Sequences/bulkUpsertAASequences
        """
        response = bulk_upsert_aa_sequences.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def find_matching_regions(self, find_matching_region: AaSequencesFindMatchingRegion) -> AsyncTaskLink:
        """
        Find matching regions for AA sequences.

        See https://benchling.com/api/v2-beta/reference#/AA%20Sequences/findMatchingRegionsAaSequences
        """
        response = find_matching_regions_aa_sequences.sync_detailed(
            client=self.client, json_body=find_matching_region
        )
        return model_from_detailed(response)
